/**
 *
 * Copyright 2004 Protique Ltd
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 **/

package org.codehaus.activemq.io.impl;
import java.io.DataInput;
import java.io.DataInputStream;
import java.io.DataOutput;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.ObjectStreamException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import javax.jms.JMSException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.codehaus.activemq.io.AbstractWireFormat;
import org.codehaus.activemq.io.WireFormat;
import org.codehaus.activemq.message.AbstractPacket;
import org.codehaus.activemq.message.CachedValue;
import org.codehaus.activemq.message.Packet;
import org.codehaus.activemq.message.WireFormatInfo;
import org.codehaus.activemq.message.util.WireByteArrayInputStream;
import org.codehaus.activemq.message.util.WireByteArrayOutputStream;

import EDU.oswego.cs.dl.util.concurrent.ConcurrentHashMap;
import EDU.oswego.cs.dl.util.concurrent.SynchronizedShort;

/**
 * Default implementation used for Java-Java protocols. When talking to non-Java nodes we may use a different wire
 * format.
 * 
 * @version $Revision: 1.12 $
 */
public class DefaultWireFormat extends AbstractWireFormat implements Serializable {

    /**
     * Current wire format version for this implementation
     */
    public static final int WIRE_FORMAT_VERSION = 3;
    private static final int MAX_CACHE_SIZE = Short.MAX_VALUE/2; //needs to be a lot less than Short.MAX_VALUE
    private static final Log log = LogFactory.getLog(DefaultWireFormat.class);
    static final short NULL_VALUE = -1;
    static final short CLEAR_CACHE = -2;
    
    private transient  PacketReader messageReader;
    private transient  PacketReader textMessageReader;
    private transient  PacketReader objectMessageReader;
    private transient  PacketReader bytesMessageReader;
    private transient  PacketReader streamMessageReader;
    private transient  PacketReader mapMessageReader;
    private transient  PacketReader messageAckReader;
    private transient  PacketReader receiptReader;
    private transient  PacketReader consumerInfoReader;
    private transient  PacketReader producerInfoReader;
    private transient  PacketReader transactionInfoReader;
    private transient  PacketReader xaTransactionInfoReader;
    private transient  PacketReader brokerInfoReader;
    private transient  PacketReader connectionInfoReader;
    private transient  PacketReader sessionInfoReader;
    private transient  PacketReader durableUnsubscribeReader;
    private transient  PacketReader reponseReceiptReader;
    private transient  PacketReader intReponseReceiptReader;
    private transient  PacketReader capacityInfoReader;
    private transient  PacketReader capacityInfoRequestReader;
    private transient  PacketReader wireFormatInfoReader;
    private transient  PacketReader keepAliveReader;
    private transient  PacketReader brokerAdminCommandReader;
    private transient  PacketReader cachedValueReader;
    private transient  PacketReader cleanupConnectionAndSessionInfoReader;        
    private transient  PacketWriter messageWriter;
    private transient  PacketWriter textMessageWriter;
    private transient  PacketWriter objectMessageWriter;
    private transient  PacketWriter bytesMessageWriter;
    private transient  PacketWriter streamMessageWriter;
    private transient  PacketWriter mapMessageWriter;
    private transient  PacketWriter messageAckWriter;
    private transient  PacketWriter receiptWriter;
    private transient  PacketWriter consumerInfoWriter;
    private transient  PacketWriter producerInfoWriter;
    private transient  PacketWriter transactionInfoWriter;
    private transient  PacketWriter xaTransactionInfoWriter;
    private transient  PacketWriter brokerInfoWriter;
    private transient  PacketWriter connectionInfoWriter;
    private transient  PacketWriter sessionInfoWriter;
    private transient  PacketWriter durableUnsubscribeWriter;
    private transient  PacketWriter reponseReceiptWriter;
    private transient  PacketWriter intReponseReceiptWriter;
    private transient  PacketWriter capacityInfoWriter;
    private transient  PacketWriter capacityInfoRequestWriter;
    private transient  PacketWriter wireFormatInfoWriter;
    private transient  PacketWriter keepAliveWriter;
    private transient  PacketWriter brokerAdminCommandWriter;
    private transient  PacketWriter cachedValueWriter;
    private transient  PacketWriter cleanupConnectionAndSessionInfoWriter;    
    private transient WireByteArrayOutputStream internalBytesOut;
    private transient DataOutputStream internalDataOut;
    private transient WireByteArrayInputStream internalBytesIn;
    private transient DataInputStream internalDataIn;
    private transient int currentWireFormatVersion;
    private transient SynchronizedShort cachedKeyGenerator;
    private transient WireByteArrayOutputStream cachedBytesOut;
    private transient DataOutputStream cachedDataOut;
    
    private List readers = new ArrayList();
    private List writers = new ArrayList();
    private Map readValueCache = new ConcurrentHashMap();
    private Map writeValueCache = new ConcurrentHashMap();
   

    /**
     * Default Constructor
     */
    public DefaultWireFormat() {
        internalBytesOut = new WireByteArrayOutputStream();
        internalDataOut = new DataOutputStream(internalBytesOut);
        internalBytesIn = new WireByteArrayInputStream();
        internalDataIn = new DataInputStream(internalBytesIn);
        this.currentWireFormatVersion = WIRE_FORMAT_VERSION;
        this.cachedKeyGenerator = new SynchronizedShort((short) 0);
        this.cachedBytesOut = new WireByteArrayOutputStream();
        this.cachedDataOut = new DataOutputStream(cachedBytesOut);
        initializeReaders();
        initializeWriters();
    }
    
    /**
     * Some wire formats require a handshake at start-up
     * @param dataOut
     * @param dataIn
     * @throws JMSException
     */
    public void initiateClientSideProtocol(DataOutputStream dataOut,DataInputStream dataIn) throws JMSException{
        WireFormatInfo info = new WireFormatInfo();
        info.setVersion(getCurrentWireFormatVersion());
        try {
            writePacket(info, dataOut);
            dataOut.flush();
        }
        catch (IOException e) {
            throw new JMSException("Failed to intiate protocol");
        }
    }
    
    
    /**
     * Some wire formats require a handshake at start-up
     * @param dataOut
     * @param dataIn
     * @throws JMSException
     */
    public void initiateServerSideProtocol(DataOutputStream dataOut,DataInputStream dataIn) throws JMSException{
    }
    
    

    /**
     * @return new WireFormat
     */
    public WireFormat copy() {
        return new DefaultWireFormat();
    }

    
    /**
     * @param dataIn
     * @return
     * @throws IOException
     */
    public Packet readPacket(DataInput dataIn) throws IOException {
        int type = -1;
        while ((type = dataIn.readByte()) == 0);
        
        if (type == -1){
            throw new IOException("InputStream now closed");
        }
        return readPacket(type, dataIn);
    }

    /**
     * @param firstByte
     * @param dataIn
     * @return
     * @throws IOException
     * 
     */
    public Packet readPacket(int firstByte, DataInput dataIn) throws IOException {
        switch (firstByte) {
            case Packet.ACTIVEMQ_MESSAGE :
                return readPacket(dataIn, messageReader);
            case Packet.ACTIVEMQ_TEXT_MESSAGE :
                return readPacket(dataIn, textMessageReader);
            case Packet.ACTIVEMQ_OBJECT_MESSAGE :
                return readPacket(dataIn, objectMessageReader);
            case Packet.ACTIVEMQ_BYTES_MESSAGE :
                return readPacket(dataIn, bytesMessageReader);
            case Packet.ACTIVEMQ_STREAM_MESSAGE :
                return readPacket(dataIn, streamMessageReader);
            case Packet.ACTIVEMQ_MAP_MESSAGE :
                return readPacket(dataIn, mapMessageReader);
            case Packet.ACTIVEMQ_MSG_ACK :
                return readPacket(dataIn, messageAckReader);
            case Packet.RECEIPT_INFO :
                return readPacket(dataIn, receiptReader);
            case Packet.CONSUMER_INFO :
                return readPacket(dataIn, consumerInfoReader);
            case Packet.PRODUCER_INFO :
                return readPacket(dataIn, producerInfoReader);
            case Packet.TRANSACTION_INFO :
                return readPacket(dataIn, transactionInfoReader);
            case Packet.XA_TRANSACTION_INFO :
                return readPacket(dataIn, xaTransactionInfoReader);
            case Packet.ACTIVEMQ_BROKER_INFO :
                return readPacket(dataIn, brokerInfoReader);
            case Packet.ACTIVEMQ_CONNECTION_INFO :
                return readPacket(dataIn, connectionInfoReader);
            case Packet.SESSION_INFO :
                return readPacket(dataIn, sessionInfoReader);
            case Packet.DURABLE_UNSUBSCRIBE :
                return readPacket(dataIn, durableUnsubscribeReader);
            case Packet.RESPONSE_RECEIPT_INFO :
                return readPacket(dataIn, reponseReceiptReader);
            case Packet.INT_RESPONSE_RECEIPT_INFO :
                return readPacket(dataIn, intReponseReceiptReader);
            case Packet.CAPACITY_INFO :
                return readPacket(dataIn, capacityInfoReader);
            case Packet.CAPACITY_INFO_REQUEST :
                return readPacket(dataIn, capacityInfoRequestReader);
            case Packet.WIRE_FORMAT_INFO :
                WireFormatInfo info =  (WireFormatInfo)readPacket(dataIn, wireFormatInfoReader);
                if (info != null){
                    if (info.getVersion() < 3){
                        throw new IOException("Cannot support wire format version: " + info.getVersion());
                    }
                }
                return info;
            
            case Packet.KEEP_ALIVE :
            	return readPacket(dataIn, keepAliveReader);
            case Packet.BROKER_ADMIN_COMMAND :
            	return readPacket(dataIn, brokerAdminCommandReader);
            case Packet.CACHED_VALUE_COMMAND :
                CachedValue cv =  (CachedValue)readPacket(dataIn,cachedValueReader);
                handleCachedValue(cv);
                return null;
            case Packet.CLEANUP_CONNECTION_INFO :
                return readPacket(dataIn, cleanupConnectionAndSessionInfoReader);
            default :
                log.error("Could not find PacketReader for packet type: "
                        + AbstractPacket.getPacketTypeAsString(firstByte));
                return null;
        }
    }

    /**
     * Write a Packet to a DataOutput
     * 
     * @param packet
     * @param dataOut
     * @return a response packet - or null
     * @throws IOException
     */
    public Packet writePacket(Packet packet, DataOutput dataOut) throws IOException {
        PacketWriter writer = getWriter(packet);
        if (writer != null) {
            validateWriteCache();
            writePacket(packet, dataOut, writer);
        }
        return null;
    }

    /**
     * A helper method which converts a packet into a byte array Overrides the WireFormat to make use of the internal
     * BytesOutputStream
     * 
     * @param packet
     * @return a byte array representing the packet using some wire protocol
     * @throws IOException
     */
    public byte[] toBytes(Packet packet) throws IOException {
        byte[] data = null;
        PacketWriter writer = getWriter(packet);
        if (writer != null) {
            internalBytesOut.reset();
            internalDataOut.writeByte(packet.getPacketType());
            internalDataOut.writeInt(-1);//the length
            writer.writePacket(packet, internalDataOut);
            internalDataOut.flush();
            data = internalBytesOut.toByteArray();
            // lets subtract the header offset from the length
            int length = data.length - 5;
            packet.setMemoryUsage(length);
            //write in the length to the data
            data[1] = (byte) ((length >>> 24) & 0xFF);
            data[2] = (byte) ((length >>> 16) & 0xFF);
            data[3] = (byte) ((length >>> 8) & 0xFF);
            data[4] = (byte) ((length >>> 0) & 0xFF);
        }
        return data;
    }
    
    /**
     * Can this wireformat process packets of this version
     * @param version the version number to test
     * @return true if can accept the version
     */
    public boolean canProcessWireFormatVersion(int version){
        return version <= WIRE_FORMAT_VERSION;
    }
    
    /**
     * @return the current version of this wire format
     */
    public int getCurrentWireFormatVersion(){
        return currentWireFormatVersion;
    }
    
    /**
     * set the current version
     * @param version
     */
    public void setCurrentWireFormatVersion(int version){
        this.currentWireFormatVersion = version;
        for (int i =0; i < readers.size(); i++){
            PacketReader reader = (PacketReader)readers.get(i);
            reader.setWireFormatVersion(version);
        }
        for (int i =0; i < writers.size(); i++){
            PacketWriter writer = (PacketWriter)writers.get(i);
            writer.setWireFormatVersion(version);
        }
    }

    protected synchronized final void writePacket(Packet packet, DataOutput dataOut, PacketWriter writer)
            throws IOException {
       
        internalBytesOut.reset();
        writer.writePacket(packet, internalDataOut);
        internalDataOut.flush();
        //reuse the byte buffer in the ByteArrayOutputStream
        byte[] data = internalBytesOut.getData();
        int count = internalBytesOut.size();
        dataOut.writeByte(packet.getPacketType());
        dataOut.writeInt(count);
        //byte[] data = internalBytesOut.toByteArray();
        //int count = data.length;
        //dataOut.writeInt(count);
        packet.setMemoryUsage(count);
        dataOut.write(data, 0, count);
    }

    protected synchronized final Packet readPacket(DataInput dataIn, PacketReader reader) throws IOException {
        Packet packet = reader.createPacket();
        int length = dataIn.readInt();
        packet.setMemoryUsage(length);
        byte[] data = new byte[length];
        dataIn.readFully(data);
        //then splat into the internal datainput
        internalBytesIn.restart(data);
        reader.buildPacket(packet, internalDataIn);
        return packet;
    }
    
    protected Object getValueFromReadCache(short key){
        return readValueCache.get(new Short(key));
    }
    
    protected short getWriteCachedKey(Object key) throws IOException{
        if (key != null){
            Short result = null;
            result = (Short)writeValueCache.get(key);
            if (result == null){
                result = new Short(cachedKeyGenerator.increment());
                writeValueCache.put(key,result);
                updateCachedValue(result.shortValue(),key);
            }
            return result.shortValue();
        }
        return DefaultWireFormat.NULL_VALUE;
    }
    
    protected void validateWriteCache() throws IOException{
        if (cachingEnabled){
        	synchronized( this ) {
	            if (writeValueCache.size() >= MAX_CACHE_SIZE){
	                writeValueCache.clear();
	                cachedKeyGenerator.set((short)0);
	                updateCachedValue((short)-1,null);//send update to peer to clear the peer cache
	            }
        	}
        }
    }

    private Object readResolve() throws ObjectStreamException {
        return new DefaultWireFormat();
    }

    private PacketWriter getWriter(Packet packet) throws IOException {
        PacketWriter answer = null;
        switch (packet.getPacketType()) {
            case Packet.ACTIVEMQ_MESSAGE :
                answer = messageWriter;
                break;
            case Packet.ACTIVEMQ_TEXT_MESSAGE :
                answer = textMessageWriter;
                break;
            case Packet.ACTIVEMQ_OBJECT_MESSAGE :
                answer = objectMessageWriter;
                break;
            case Packet.ACTIVEMQ_BYTES_MESSAGE :
                answer = bytesMessageWriter;
                break;
            case Packet.ACTIVEMQ_STREAM_MESSAGE :
                answer = streamMessageWriter;
                break;
            case Packet.ACTIVEMQ_MAP_MESSAGE :
                answer = mapMessageWriter;
                break;
            case Packet.ACTIVEMQ_MSG_ACK :
                answer = messageAckWriter;
                break;
            case Packet.RECEIPT_INFO :
                answer = receiptWriter;
                break;
            case Packet.CONSUMER_INFO :
                answer = consumerInfoWriter;
                break;
            case Packet.PRODUCER_INFO :
                answer = producerInfoWriter;
                break;
            case Packet.TRANSACTION_INFO :
                answer = transactionInfoWriter;
                break;
            case Packet.XA_TRANSACTION_INFO :
                answer = xaTransactionInfoWriter;
                break;
            case Packet.ACTIVEMQ_BROKER_INFO :
                answer = brokerInfoWriter;
                break;
            case Packet.ACTIVEMQ_CONNECTION_INFO :
                answer = connectionInfoWriter;
                break;
            case Packet.SESSION_INFO :
                answer = sessionInfoWriter;
                break;
            case Packet.DURABLE_UNSUBSCRIBE :
                answer = durableUnsubscribeWriter;
                break;
            case Packet.RESPONSE_RECEIPT_INFO :
                answer = reponseReceiptWriter;
                break;
            case Packet.INT_RESPONSE_RECEIPT_INFO :
                answer = intReponseReceiptWriter;
                break;
            case Packet.CAPACITY_INFO :
                answer = capacityInfoWriter;
                break;
            case Packet.CAPACITY_INFO_REQUEST :
                answer = capacityInfoRequestWriter;
                break;
            case Packet.WIRE_FORMAT_INFO :
                answer = wireFormatInfoWriter;
                break;
            case Packet.KEEP_ALIVE :
                answer = keepAliveWriter;
                break;
            case Packet.BROKER_ADMIN_COMMAND :
                answer = brokerAdminCommandWriter;
                break;
            case Packet.CACHED_VALUE_COMMAND:
                answer = cachedValueWriter;
                break;
            case Packet.CLEANUP_CONNECTION_INFO:
                answer = cleanupConnectionAndSessionInfoWriter;
                break;
            default :
                log.error("no PacketWriter for packet: " + packet);
        }
        return answer;
    }
    
    private void handleCachedValue(CachedValue cv){
        if (cv != null){
            if (cv.getId() == CLEAR_CACHE){
                //clear the cache
                readValueCache.clear();
            }else if (cv.getId() != NULL_VALUE){
                readValueCache.put(new Short(cv.getId()), cv.getValue());
            }
        }
    }
    
    private void initializeReaders() {
        messageReader = new ActiveMQMessageReader(this);
        readers.add(messageReader);
        textMessageReader = new ActiveMQTextMessageReader(this);
        readers.add(textMessageReader);
        objectMessageReader = new ActiveMQObjectMessageReader(this);
        readers.add(objectMessageReader);
        bytesMessageReader = new ActiveMQBytesMessageReader(this);
        readers.add(bytesMessageReader);
        streamMessageReader = new ActiveMQStreamMessageReader(this);
        readers.add(streamMessageReader);
        mapMessageReader = new ActiveMQMapMessageReader(this);
        readers.add(mapMessageReader);
        messageAckReader = new MessageAckReader(this);
        readers.add(messageAckReader);
        receiptReader = new ReceiptReader();
        readers.add(receiptReader);
        consumerInfoReader = new ConsumerInfoReader();
        readers.add(consumerInfoReader);
        producerInfoReader = new ProducerInfoReader();
        readers.add(producerInfoReader);
        transactionInfoReader = new TransactionInfoReader();
        readers.add(transactionInfoReader);
        xaTransactionInfoReader = new XATransactionInfoReader();
        readers.add(xaTransactionInfoReader);
        brokerInfoReader = new BrokerInfoReader();
        readers.add(brokerInfoReader);
        connectionInfoReader = new ConnectionInfoReader();
        readers.add(connectionInfoReader);
        sessionInfoReader = new SessionInfoReader();
        readers.add(sessionInfoReader);
        durableUnsubscribeReader = new DurableUnsubscribeReader();
        readers.add(durableUnsubscribeReader);
        reponseReceiptReader = new ResponseReceiptReader();
        readers.add(reponseReceiptReader);
        intReponseReceiptReader = new IntResponseReceiptReader();
        readers.add(intReponseReceiptReader);
        capacityInfoReader = new CapacityInfoReader();
        readers.add(capacityInfoReader);
        capacityInfoRequestReader = new CapacityInfoRequestReader();
        readers.add(capacityInfoReader);
        wireFormatInfoReader = new WireFormatInfoReader(this);
        readers.add(wireFormatInfoReader);
        keepAliveReader = new KeepAliveReader();
        readers.add(keepAliveReader);
        brokerAdminCommandReader = new BrokerAdminCommandReader();
        readers.add(brokerAdminCommandReader);
        cachedValueReader = new CachedValueReader();
        readers.add(cachedValueReader);
        cleanupConnectionAndSessionInfoReader = new CleanupConnectionInfoReader();
        readers.add(cleanupConnectionAndSessionInfoReader);
    }
    
    private void initializeWriters(){
        messageWriter = new ActiveMQMessageWriter(this);
        writers.add(messageWriter);
        textMessageWriter = new ActiveMQTextMessageWriter(this);
        writers.add(textMessageWriter);
        objectMessageWriter = new ActiveMQObjectMessageWriter(this);
        writers.add(objectMessageWriter);
        bytesMessageWriter = new ActiveMQBytesMessageWriter(this);
        writers.add(bytesMessageWriter);
        streamMessageWriter = new ActiveMQStreamMessageWriter(this);
        writers.add(streamMessageWriter);
        mapMessageWriter = new ActiveMQMapMessageWriter(this);
        writers.add(mapMessageWriter);
        messageAckWriter = new MessageAckWriter(this);
        writers.add(messageAckWriter);
        receiptWriter = new ReceiptWriter();
        writers.add(receiptWriter);
        consumerInfoWriter = new ConsumerInfoWriter();
        writers.add(consumerInfoWriter);
        producerInfoWriter = new ProducerInfoWriter();
        writers.add(producerInfoWriter);
        transactionInfoWriter = new TransactionInfoWriter();
        writers.add(transactionInfoWriter);
        xaTransactionInfoWriter = new XATransactionInfoWriter();
        writers.add(xaTransactionInfoWriter);
        brokerInfoWriter = new BrokerInfoWriter();
        writers.add(brokerInfoWriter);
        connectionInfoWriter = new ConnectionInfoWriter();
        writers.add(connectionInfoWriter);
        sessionInfoWriter = new SessionInfoWriter();
        writers.add(sessionInfoWriter);
        durableUnsubscribeWriter = new DurableUnsubscribeWriter();
        writers.add(durableUnsubscribeWriter);
        reponseReceiptWriter = new ResponseReceiptWriter();
        writers.add(reponseReceiptWriter);
        intReponseReceiptWriter = new IntResponseReceiptWriter();
        writers.add(intReponseReceiptWriter);
        capacityInfoWriter = new CapacityInfoWriter();
        writers.add(capacityInfoWriter);
        capacityInfoRequestWriter = new CapacityInfoRequestWriter();
        writers.add(capacityInfoWriter);
        wireFormatInfoWriter = new WireFormatInfoWriter();
        writers.add(wireFormatInfoWriter);
        keepAliveWriter = new KeepAliveWriter();
        writers.add(keepAliveWriter);
        brokerAdminCommandWriter = new BrokerAdminCommandWriter();
        writers.add(brokerAdminCommandWriter);
        cachedValueWriter = new CachedValueWriter();
        writers.add(cachedValueWriter);
        cleanupConnectionAndSessionInfoWriter = new CleanupConnectionInfoWriter();
        writers.add(cleanupConnectionAndSessionInfoWriter);
    }
    
    private synchronized void updateCachedValue(short key,Object value) throws IOException{
        if (cachedValueWriter == null){
            cachedValueWriter = new CachedValueWriter();
        }
        CachedValue cv = new CachedValue();
        cv.setId(key);
        cv.setValue(value);
        cachedBytesOut.reset();
        cachedValueWriter.writePacket(cv,cachedDataOut);
        cachedDataOut.flush();
        byte[] data = cachedBytesOut.getData();
        int count = cachedBytesOut.size();
        getTransportDataOut().writeByte(cv.getPacketType());
        getTransportDataOut().writeInt(count);
        getTransportDataOut().write(data,0,count);
        
        
    }
}