/** 
 * 
 * Copyright 2004 Protique Ltd
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License. 
 * 
 **/

package org.codehaus.activemq.message.util;

import org.codehaus.activemq.message.ActiveMQMessage;
import org.codehaus.activemq.service.QueueListEntry;
import org.codehaus.activemq.service.impl.DefaultQueueList;

import javax.jms.JMSException;
import java.util.ArrayList;
import java.util.List;

/**
 * MemoryBoundedQueue is a queue bounded by memory usage for ActiveMQMessages
 *
 * @version $Revision: 1.5 $
 */
public class MemoryBoundedQueue implements BoundedActiveMQMessageQueue {
    private static final int WAIT_TIMEOUT = 100;

    private MemoryBoundedQueueManager queueManager;
    private String name;
    private boolean stopped = false;
    private boolean closed = false;
    private long memoryUsedByThisQueue;
    private Object outLock = new Object();
    private Object inLock = new Object();
    private DefaultQueueList internalList = new DefaultQueueList();

    /**
     * Constructor
     *
     * @param name
     * @param manager
     */
    MemoryBoundedQueue(String name, MemoryBoundedQueueManager manager) {
        this.name = name;
        this.queueManager = manager;
    }

    /**
     * @return the name of this MemoryBoundedQueue
     */
    public String getName() {
        return name;
    }

    /**
     * @return a pretty print of this queue
     */
    public String toString() {
        return "" + name + " , cardinality = " + size() + " memory usage = " + memoryUsedByThisQueue + " , contents{" + internalList +"}";
    }

    /**
     * @return the number of items held by this queue
     */
    public int size() {
        return internalList.size();
    }

    /**
     * @return an aproximation the memory used by this queue
     */
    public long getLocalMemoryUsedByThisQueue() {
        return memoryUsedByThisQueue;
    }

    /**
     * close and remove this queue from the MemoryBoundedQueueManager
     */
    public void close() {
        try {
            clear();
            closed = true;
            synchronized (outLock) {
                outLock.notifyAll();
            }
            synchronized (inLock) {
                inLock.notifyAll();
            }
        }
        catch (Throwable e) {
            e.printStackTrace();
        }
        finally {
            queueManager.removeMemoryBoundedQueue(getName());
        }
    }

    /**
     * Enqueue a ActiveMQMessage without checking memory usage limits
     *
     * @param packet
     */
    public void enqueueNoBlock(ActiveMQMessage packet) {
        if (!closed) {
            internalList.add(packet);
            incrementMemoryUsed(packet);
            synchronized (outLock) {
                outLock.notify();
            }
        }
    }

    /**
     * Enqueue a ActiveMQMessage to this queue
     *
     * @param packet
     */
    public void enqueue(ActiveMQMessage packet) {
        if (!queueManager.isFull()) {
            enqueueNoBlock(packet);
        }
        else {
            synchronized (inLock) {
                try {
                    while (queueManager.isFull() && !closed) {
                        inLock.wait(WAIT_TIMEOUT);
                    }
                }
                catch (InterruptedException ie) {
                }
            }
            enqueueNoBlock(packet);
        }
    }

    /**
     * Enqueue a packet to the head of the queue with total disregard for memory constraints
     *
     * @param packet
     */
    public final void enqueueFirstNoBlock(ActiveMQMessage packet) {
        if (!closed) {
            internalList.addFirst(packet);
            incrementMemoryUsed(packet);
            synchronized (outLock) {
                outLock.notify();
            }
        }
    }

    /**
     * Enqueue a ActiveMQMessage to the head of the queue
     *
     * @param packet
     * @throws InterruptedException
     */
    public void enqueueFirst(ActiveMQMessage packet) throws InterruptedException {
        if (!queueManager.isFull()) {
            enqueueFirstNoBlock(packet);
        }
        else {
            synchronized (inLock) {
                while (queueManager.isFull() && !closed) {
                    inLock.wait(WAIT_TIMEOUT);
                }
            }
            enqueueFirstNoBlock(packet);
        }
    }

    /**
     * @return the first dequeued ActiveMQMessage or blocks until one is available
     * @throws InterruptedException
     */
    public ActiveMQMessage dequeue() throws InterruptedException {
        ActiveMQMessage result = null;
        synchronized (outLock) {
            while (internalList.isEmpty() && !closed) {
                outLock.wait(WAIT_TIMEOUT);
            }
            result = dequeueNoWait();
        }
        return result;
    }

    /**
     * Dequeues a ActiveMQMessage from the head of the queue
     *
     * @param timeInMillis time to wait for a ActiveMQMessage to be available
     * @return the first ActiveMQMessage or null if none available within <I>timeInMillis </I>
     * @throws InterruptedException
     */
    public ActiveMQMessage dequeue(long timeInMillis) throws InterruptedException {
        ActiveMQMessage result = null;
        if (timeInMillis == 0) {
            result = dequeue();
        }
        else {
            synchronized (outLock) {
                // if timeInMillis is less than zero assume nowait
                long waitTime = timeInMillis;
                long start = (timeInMillis <= 0) ? 0 : System.currentTimeMillis();
                while (!closed) {
                    result = dequeueNoWait();
                    if (result != null || waitTime <= 0) {
                        break;
                    }
                    else {
                        outLock.wait(waitTime);
                        waitTime = timeInMillis - (System.currentTimeMillis() - start);
                    }
                }
            }
        }
        return result;
    }

    /**
     * dequeues a ActiveMQMessage from the head of the queue
     *
     * @return the ActiveMQMessage at the head of the queue or null, if none is available
     * @throws InterruptedException
     */
    public ActiveMQMessage dequeueNoWait() throws InterruptedException {
        ActiveMQMessage packet = null;
        synchronized (outLock) {
            while (stopped && !closed) {
                outLock.wait(WAIT_TIMEOUT);
            }
        }
        packet = (ActiveMQMessage) internalList.removeFirst();
        decrementMemoryUsed(packet);
        if (packet != null) {
            synchronized (inLock) {
                inLock.notify();
            }
        }
        return packet;
    }

    /**
     * @return true if the queue is enabled for dequeing (default = true)
     */
    public boolean isStarted() {
        synchronized (outLock) {
            return stopped == false;
        }
    }

    /**
     * disable dequeueing
     */
    public void stop() {
        synchronized (outLock) {
            stopped = true;
        }
    }

    /**
     * enable dequeueing
     */
    public void start() {
        synchronized (outLock) {
            stopped = false;
            outLock.notifyAll();
        }
        synchronized (inLock) {
            inLock.notifyAll();
        }
    }

    /**
     * Remove a packet from the queue
     *
     * @param packet
     * @return true if the packet was found
     */
    public boolean remove(ActiveMQMessage packet) {
        boolean result = false;
        if (!internalList.isEmpty()) {
            result = internalList.remove(packet);
        }
        if (result) {
            decrementMemoryUsed(packet);
        }
        synchronized (inLock) {
            inLock.notify();
        }
        return result;
    }

    /**
     * Remove a ActiveMQMessage by it's id
     *
     * @param id
     * @return
     */
    public ActiveMQMessage remove(String id) {
        ActiveMQMessage result = null;
        QueueListEntry entry = internalList.getFirstEntry();
        try {
            while (entry != null) {
                ActiveMQMessage p = (ActiveMQMessage) entry.getElement();
                if (p.getJMSMessageID().equals(id)) {
                    result = p;
                    remove(p);
                    break;
                }
                entry = internalList.getNextEntry(entry);
            }
        }
        catch (JMSException jmsEx) {
            jmsEx.printStackTrace();
        }
        synchronized (inLock) {
            inLock.notify();
        }
        return result;
    }

    /**
     * remove any ActiveMQMessages in the queue
     */
    public void clear() {
        while (!internalList.isEmpty()) {
            ActiveMQMessage packet = (ActiveMQMessage) internalList.removeFirst();
            decrementMemoryUsed(packet);
        }
        synchronized (inLock) {
            inLock.notifyAll();
        }
    }

    /**
     * @return true if the queue is empty
     */
    public boolean isEmpty() {
        return internalList.isEmpty();
    }

    /**
     * retrieve a ActiveMQMessage at an indexed position in the queue
     *
     * @param index
     * @return
     */
    public ActiveMQMessage get(int index) {
        return (ActiveMQMessage) internalList.get(index);
    }

    /**
     * Retrieve a shallow copy of the contents as a list
     *
     * @return a list containing the bounded queue contents
     */
    public List getContents() {
        Object[] array = internalList.toArray();
        List list = new ArrayList();
        for (int i = 0; i < array.length; i++) {
            list.add(array[i]);
        }
        return list;
    }

    private synchronized void incrementMemoryUsed(ActiveMQMessage packet) {
        if (packet != null) {
            memoryUsedByThisQueue += queueManager.incrementMemoryUsed(packet);
        }
    }

    private synchronized void decrementMemoryUsed(ActiveMQMessage packet) {
        if (packet != null) {
            memoryUsedByThisQueue -= queueManager.decrementMemoryUsed(packet);
        }
    }
}